import fnmatch
import re
import logging

from typing import Union
from collections import Iterable
from pygtrie import PrefixSet
from urllib.parse import urlparse, ParseResult
from bs4 import BeautifulSoup

from cobweb.link_extractor.bloom_filter import BloomFilter


class LinkExtractor:

    REGEX_UNIX = "unix"
    REGEX_REGULAR = "regular"

    PROBABILISTIC = "probabilistic"
    TRIE = "trie"
    HASH = "hash"

    def __init__(
            self,
            follow: Union[Iterable, str] = ("*",),
            exclude: Union[Iterable, str] = tuple(),
            regex_type=REGEX_UNIX,
            mode=HASH,
            domain_only=True
    ):

        self._current_page_url = None
        self._follow_patterns = None
        self._exclude_patterns = None

        self.domain_only = domain_only
        self.allowed_schemes = {
            "http",
            "https",
        }

        self.regex_type = regex_type
        self.follow_patterns = follow
        self.exclude_patterns = exclude

        self.mode = mode
        # Any object with both '__contains__' and 'add' methods defined can be used here
        if mode == LinkExtractor.PROBABILISTIC:
            self.seen = BloomFilter()
        elif mode == LinkExtractor.HASH:
            self.seen = set()
        elif mode == LinkExtractor.TRIE:
            self.seen = PrefixSet()

    @property
    def logger(self):
        return logging.getLogger("cobweb.link_extractor")

    @property
    def current_page_url(self):
        return self._current_page_url

    @current_page_url.setter
    def current_page_url(self, value: str):
        self._current_page_url = urlparse(value)

    @property
    def follow_patterns(self):
        return self._follow_patterns

    @follow_patterns.setter
    def follow_patterns(self, value):
        follow_patterns = value if not isinstance(value, str) else [value]
        self._follow_patterns = [
            self.get_regex(pattern)
            for pattern in follow_patterns
        ]

    @property
    def exclude_patterns(self):
        return self._exclude_patterns

    @exclude_patterns.setter
    def exclude_patterns(self, value):
        exclude_patterns = value if not isinstance(value, str) else [value]
        self.exclude_patterns = [
            self.get_regex(pattern)
            for pattern in exclude_patterns
        ]

    def get_regex(self, string: str):
        return re.compile(fnmatch.translate(string)) if self.regex_type == self.REGEX_UNIX else re.compile(string)

    def is_valid_domain(self, href: str) -> bool:
        """
        checks whether the given string satisfy the 'domain_only' setting
        :param href: str
        :return: bool
        """
        parsed = urlparse(href)
        if parsed.scheme and parsed.scheme not in self.allowed_schemes:
            return False
        if self.domain_only:
            return not parsed.netloc or parsed.netloc == self.current_page_url.netloc
        else:
            return parsed.scheme and parsed.netloc

    def match_follow_patterns(self, link):
        return any([
                pattern.match(link)
                for pattern in self.follow_patterns
        ])

    def match_exclude_patterns(self, link):
        return any([
            pattern.match(link)
            for pattern in self.exclude_patterns
        ])

    def is_satisfying_link(self, link: str) -> bool:
        """
        checks whether the given link satisfy 'follow' and 'exclude' preferences
        :param link: str
        :return: bool
        """
        return self.match_follow_patterns(link) and not self.match_exclude_patterns(link)

    def get_full_url(self, href: ParseResult):
        if not href.scheme:
            href = href._replace(scheme=self.current_page_url.scheme)
        if not href.netloc:
            href = href._replace(netloc=self.current_page_url.netloc)
        if href.path.startswith("./"):
            href = href._replace(path=href.path[2:])
        return href.geturl()

    def filter_valid_links(self, all_links: set):
        valid_links = set()
        for link in all_links:
            if link.has_attr("href") and self.is_valid_domain(link["href"]):
                url = self.get_full_url(urlparse(link["href"]))
                if self.is_satisfying_link(url):
                    valid_links.add(url)
                    self.seen.add(url)
        self.logger.debug("valid links count: %s", len(valid_links))
        if self.mode == self.PROBABILISTIC:
            self.logger.info("current error probability %s", self.seen.count_error_probability())
        return valid_links

    def extract(self, response):
        self.current_page_url = response.url
        data = BeautifulSoup(response.content, "html.parser")
        all_links = data.find_all("a")
        self.logger.debug("found %s new links", len(all_links))
        return self.filter_valid_links(all_links)



import requests

MAX_FILTER_SIZE = 1e7


def error_callback_pass():
    pass


def error_callback_raise():
    raise requests.HTTPError("Bad return code")


def queue_timeout_wait():
    print("here"*80)


ERROR_CALLBACKS = {
    "skip": error_callback_pass,
    "raise": error_callback_raise,

}

QUEUE_TIMEOUT_CALLBACKS = {
    "wait": queue_timeout_wait
}



from abc import abstractmethod, ABCMeta


class AbstractAdapter(metaclass=ABCMeta):

    @abstractmethod
    def __call__(self, *args, **kwargs):
        raise NotImplementedError("Method invocation on an abstract class")

    @abstractmethod
    def prepare_request(self, *args, **kwargs):
        raise NotImplementedError("Method invocation on an abstract class")

    @abstractmethod
    def process_response(self, response, *args, **kwargs):
        raise NotImplementedError("Method invocation on an abstract class")



import logging

from requests import Request, Session
from requests.cookies import RequestsCookieJar
from requests.exceptions import HTTPError

from cobweb.adapters import AbstractAdapter


class Http(AbstractAdapter):

    def __init__(
            self,
            method="GET",
            stream=False,
            proxies=None,
            ssl_certificate=None,
            timeout=(10, 10),
            data=None,
            headers=None):

        self._headers = headers
        self._data = data
        self._timeout = timeout
        self._ssl_certificate = ssl_certificate
        self._proxies = proxies
        self._method = method
        self._stream = stream
        self._response = None
        self._content = None
        self.session = None
        self.accept_content_types = {
            "text/html",
            "text/plain",
            "application/xhtml+xml",
            "application/xml",
        }

    @property
    def logger(self):
        return logging.getLogger("cobweb.adapters.http")

    @property
    def headers(self) -> dict:
        return self._headers

    @property
    def cookies(self) -> RequestsCookieJar:
        return self._cookies

    def __enter__(self):
        self.session = Session()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_session()

    def __str__(self) -> str:
        return "Http response status {}: \n{}".format(self.status_code, self.content.body)

    def __call__(self, url: str):
        if not self.session:
            self.session = Session()
            response = self.send(url)
            self.close_session()
        else:
            response = self.send(url)
        return response

    def close_session(self):
        self.session.close()
        self.session = None

    def send(self, url: str):
        self.logger.debug("open session")
        with self.session as s:
            response = s.send(
                self.prepare_request(url),
                stream=self._stream,
                verify=self._ssl_certificate,
                proxies=self._proxies,
                cert=self._ssl_certificate,
                timeout=self._timeout
            )
        self.logger.debug("got response")
        return self.process_response(response)

    def prepare_request(self, url: str):
        self.logger.debug("preparing request")
        request = Request(self._method, url, data=self._data, headers=self._headers)
        return self.session.prepare_request(request)

    def process_response(self, response):
        if response.headers["content-type"].partition(";")[0] not in self.accept_content_types:
            raise HTTPError("Wrong content type")
        return response




from distutils.core import setup

setup(
    name='cobweb_scrapper',
    version='0.1.0',
    packages=['cobweb', 'cobweb.spiders',
              'cobweb.adapters', 'cobweb.link_extractor'],
    url='https://bitbucket.org/pollux_cast0r/cobweb/overview',
    license='',
    author='Alexander Svito',
    author_email='alexandervirk@gmail.com',
    description='A package for easy scrapping', requires=['requests', 'PyYAML', 'pygtrie', 'bs4', 'urllib3']
)



import multiprocessing
import functools
import requests


def run_in_another_process(func):
    @functools.wraps(func)
    def inner(*args, **kwargs):
        p = multiprocessing.Process(target=func, args=args, kwargs=kwargs)
        p.start()
    return inner


@run_in_another_process
def extract_links(queue, link_extractor, response):
    for link in link_extractor.extract(response):
        queue.put(link)


def is_connected_to_internet():
    try:
        requests.get('http://216.58.192.142', timeout=2)
        return True
    except requests.exceptions.ConnectionError:
        return False


class Singleton(type):

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]

